local MAJOR, MINOR = "GetItemInfoAsync-1.0", 0
local LIBRARY = LibStub:NewLibrary(MAJOR, MINOR)
if not LIBRARY then return end -- Already loaded


---------------------------------------------
-- DISPATCHER
---------------------------------------------
local function errorhandler(err)
    return geterrorhandler()(err)
end

-- We use a runtime compiled function so we can give it a specifc name in the callstack
local dispatcher_code = [[
    local xpcall, unpack, errorhandler = ...

    return function(callbacks, ...)
        local args = { ... }

        local i, callback = next(callbacks)
        local function invoke() callback(unpack(args)) end
        while callback do
            xpcall(invoke, errorhandler)
            i, callback = next(callbacks, i)
        end
    end
]]
local dispatcher = assert(loadstring(dispatcher_code, "safecall GetItemInfoAsync:Dispatch"))(xpcall, unpack, errorhandler)


---------------------------------------------
-- REGISTRY
---------------------------------------------
local Registry = {
    New = function(self, onFirst, onLast)
        local obj = {
            callbacksByEvent = {},
            count = 0,
            ["onFirst"] = onFirst,
            ["onLast"] = onLast
        }
        setmetatable(obj, self)
        return obj
    end;

    Register = function(self, eventname, callback)
        if ( self.onFirst and self.count <= 0 ) then
            self.onFirst()
        end

        if ( not self.callbacksByEvent[eventname] ) then
            self.callbacksByEvent[eventname] = {}
        end
        tinsert(self.callbacksByEvent[eventname], callback)
        self.count = self.count + 1
    end;

    Unregister = function(self, eventname)
        local callbacks = self.callbacksByEvent[eventname]
        if ( callbacks ) then
            self.callbacksByEvent[eventname] = nil
            self.count = self.count - #callbacks

            if ( self.onLast and self.count <= 0 ) then
                self.onLast()
            end
        end
    end;

    Dispatch = function(self, eventname, ...)
        local callbacks = self.callbacksByEvent[eventname]
        if ( callbacks ) then
            dispatcher(callbacks, ...)
        end
    end;
}
Registry.__index = Registry


---------------------------------------------
-- INITIALIZE LIBRARY
---------------------------------------------
local frame = CreateFrame("FRAME")
local CallbacksByItem = Registry:New(
    function() frame:RegisterEvent("GET_ITEM_INFO_RECEIVED") end, --[[onFirst]]
    function() frame:UnregisterEvent("GET_ITEM_INFO_RECEIVED") end --[[onLast]])

frame:Hide()
frame:SetScript("OnEvent", function(self, event, itemID, success)
    if ( event == "GET_ITEM_INFO_RECEIVED" and success ) then
        local itemInfo = { GetItemInfo(itemID) }
        local itemName = itemInfo[1]

        -- Dispatch to itemID
        CallbacksByItem:Dispatch(itemID, unpack(itemInfo))
        CallbacksByItem:Unregister(itemID)
        -- Dispatch to itemName
        CallbacksByItem:Dispatch(itemName, unpack(itemInfo))
        CallbacksByItem:Unregister(itemName)
    end
end)

local function GetItemInfoAsync(self, item, callback)
    -- Extract itemID from itemLink and itemString
    if ( type(item) == "string" and item:match("item:%d+") ) then
        item = tonumber(item:match("item:(%d+)"))
    end

    local info = { GetItemInfo(item) }
    -- Check to see if the item info is already available.
    if ( info[1] ) then
        callback(unpack(info))
        return true
    else
        CallbacksByItem:Register(item, callback)
        return false
    end
end
setmetatable(LIBRARY, { __call = GetItemInfoAsync })


---------------------------------------------
-- TESTS
---------------------------------------------
--@alpha@
-- Test single dispatch
local testvalue = 0
local testcallbacks = {
    [1] = function(msg)
        assert(msg == "Hello world", "TEST_FAILED Single Dispatch. Unexpected message")
        testvalue = testvalue + 1
    end
}
dispatcher(testcallbacks, "Hello world")
assert(testvalue == 1, "TEST_FAILED Single Dispatch. Unexpected testvalue")
-- Test multiple dispatch
testvalue = 0
testcallbacks = {
    [1] = function(msg)
        assert(msg == "Hello world", "TEST_FAILED Multiple Dispatch. Unexpected message")
        testvalue = testvalue + 1
    end,
    [2] = function(msg)
        assert(msg == "Hello world", "TEST_FAILED Multiple Dispatch. Unexpected message")
        testvalue = testvalue + 2
    end
}
dispatcher(testcallbacks, "Hello world")
assert(testvalue == 3, "TEST_FAILED Multiple Dispatch. Unexpected testvalue")
-- Test noncontiguous dispatch
testvalue = 0
testcallbacks = {
    [1] = function(msg)
        assert(msg == "Hello world", "TEST_FAILED NonContiguous Dispatch. Unexpected message")
        testvalue = testvalue + 1
    end,
    [5] = function(msg)
        assert(msg == "Hello world", "TEST_FAILED NonContiguous Dispatch. Unexpected message")
        testvalue = testvalue + 2
    end
}
dispatcher(testcallbacks, "Hello world")
assert(testvalue == 3, "TEST_FAILED NonContiguous Dispatch. Unexpected testvalue")
-- Test mixed dispatch
testvalue = 0
testcallbacks = {
    [1] = function(msg)
        assert(msg == "Hello world", "TEST_FAILED Mixed Dispatch. Unexpected message")
        testvalue = testvalue + 1
    end,
    [2] = function(msg)
        assert(msg == "Hello world", "TEST_FAILED Mixed Dispatch. Unexpected message")
        testvalue = testvalue + 2
    end,
    ["tree"] = function(msg)
        assert(msg == "Hello world", "TEST_FAILED Mixed Dispatch. Unexpected message")
        testvalue = testvalue + 3
    end
}
dispatcher(testcallbacks, "Hello world")
assert(testvalue == 6, "TEST_FAILED Multiple Dispatch. Unexpected testvalue")

-- Test Registry
local testregistry = Registry:New()
testvalue = 0
testregistry:Dispatch("testevent", "Unexpected") -- No callbacks
testregistry:Register("testevent", function(msg)
    assert(msg == "Hello world", "TEST_FAILED Registry. Unexpected message")
    testvalue = testvalue + 1
end)
testregistry:Register("testevent", function(msg)
    assert(msg == "Hello world", "TEST_FAILED Registry. Unexpected message")
    testvalue = testvalue + 2
end)
testregistry:Register("othertestevent", function(msg)
print("othertestevent")
    error("TEST_FAILED Registry. Invoked othertestevent callback")
end)
testregistry:Dispatch("testevent", "Hello world")
assert(testvalue == 3, "TEST_FAILED Registry. Unexpected testvalue")
testregistry:Unregister("testevent")
testregistry:Dispatch("testevent", "Unexpected") -- No callbacks
assert(testvalue == 3, "TEST_FAILED Registry. No callbacks should have been invoked")

-- Test Registry with onFirst & onLast
local first, last = 0, 0
testregistry = Registry:New(
    function() first = first + 1 end,
    function() last = last + 1 end)
testregistry:Dispatch("testevent") -- No callbacks
testregistry:Register("testevent", function(msg) end)
assert(first == 1 , "TEST_FAILED Registry with First/Last. Registered testevent callback 1, expected first == 1")
assert(last == 0 , "TEST_FAILED Registry with First/Last. Registered testevent callback 1, expected last == 0")
testregistry:Register("testevent", function(msg) end)
assert(first == 1 , "TEST_FAILED Registry with First/Last. Registered testevent callback 2, expected first == 1")
assert(last == 0 , "TEST_FAILED Registry with First/Last. Registered testevent callback 2, expected last == 0")
testregistry:Register("othertestevent", function(msg) end)
assert(first == 1 , "TEST_FAILED Registry with First/Last. Registered othertestevent callback, expected first == 1")
assert(last == 0 , "TEST_FAILED Registry with First/Last. Registered othertestevent callback, expected last == 0")
testregistry:Dispatch("testevent")
testregistry:Unregister("testevent")
assert(first == 1, "TEST_FAILED Registry with First/Last. Unregister testevent callbacks, expected first == 1")
assert(last == 0, "TEST_FAILED Registry with First/Last. Unregister testevent callbacks, expected last == 0")
testregistry:Unregister("othertestevent")
assert(first == 1, "TEST_FAILED Registry with First/Last. Unregister othertestevent callbacks, expected first == 1")
assert(last == 1, "TEST_FAILED Registry with First/Last. Unregister othertestevent callbacks, expected last == 1")
testregistry:Dispatch("testevent") -- No callbacks
testregistry = nil

-- Test GetItemInfoAsync(itemID)
local testeditemid = false
LIBRARY(12468, function(itemName, itemLink) -- "Chilton Wand"
    testeditemid = true
    assert(tonumber(itemLink:match("item:(%d+)")) == 12468, "TEST_FAILED GetItemInfoAsynch(itemID). Unexpected itemId")
end)
C_Timer.After(1, function() assert(testeditemid, "TEST_FAILED GetItemInfoAsync(itemID). Was not invoked") end)
-- Test GetItemInfoAsync(itemName)
local testitemlink = nil
LIBRARY("Hearthstone", function(itemName, itemLink) -- "Hearthstone"
    testitemname = true
    assert(tonumber(itemLink:match("item:(%d+)")) == 6948, "TEST_FAILED GetItemInfoAsynch(itemName). Unexpected itemId")
end)
C_Timer.After(1, function() assert(testeditemid, "TEST_FAILED GetItemInfoAsynch(itemName). Was not invoked") end)
-- Test GetItemInfoAsync(itemString)
local testitemstring = false
LIBRARY("item:18832:2564:0:0:0:0:0:0:80:0:0:0:0", function(itemName, itemLink) -- "Brutality Blade"
    testitemstring = true
    assert(tonumber(itemLink:match("item:(%d+)")) == 18832, "TEST_FAILED GetItemInfoAsynch(itemString). Unexpected itemId")
end)
C_Timer.After(1, function() assert(testitemstring, "TEST_FAILED GetItemInfoAsynch(itemString). Was not invoked") end)
-- Test GetItemInfoAsync(itemLink)
LIBRARY(6948, function(itemName, itemLink) -- "Hearthstone"
    local testitemlink = false
    LIBRARY(itemLink, function(itemName, itemLink)
        testitemlink = true
        assert(tonumber(itemLink:match("item:(%d+)")) == 6948, "TEST_FAILED GetItemInfoAsynch(itemLink). Unexpected itemId")
    end)
    C_Timer.After(1, function() assert(testitemlink, "TEST_FAILED GetItemInfoAsynch(itemString). Was not invoked") end)
end)
--@end-alpha@
